@@ -1,14 +1,17 @@
 import copy
 import collections
+import os
 import sys
-from nose.tools import assert_equal, assert_raises, assert_true
+
+from nose.tools import assert_equal, assert_false, assert_raises, assert_true
 
 from voluptuous import (
-    Schema, Required, Optional, Extra, Invalid, In, Remove, Literal,
-    Url, MultipleInvalid, LiteralInvalid, NotIn, Match, Email,
+    Schema, Required, Exclusive, Optional, Extra, Invalid, In, Remove, Literal,
+    Url, MultipleInvalid, LiteralInvalid, TypeInvalid, NotIn, Match, Email,
     Replace, Range, Coerce, All, Any, Length, FqdnUrl, ALLOW_EXTRA, PREVENT_EXTRA,
     validate, ExactSequence, Equal, Unordered, Number, Maybe, Datetime, Date,
-    Contains, Marker)
+    Contains, Marker, IsDir, IsFile, PathExists, SomeOf, TooManyValid, Self,
+    raises)
 from voluptuous.humanize import humanize_error
 from voluptuous.util import u
 
@@ -153,6 +156,39 @@ def test_literal():
         assert False, "Did not raise Invalid"
 
 
+def test_class():
+    class C1(object):
+        pass
+
+    schema = Schema(C1)
+    schema(C1())
+
+    try:
+        schema(None)
+    except MultipleInvalid as e:
+        assert_equal(str(e), "expected C1")
+        assert_equal(len(e.errors), 1)
+        assert_equal(type(e.errors[0]), TypeInvalid)
+    else:
+        assert False, "Did not raise Invalid"
+
+    # In Python 2, this will be an old-style class (classobj instance)
+    class C2:
+        pass
+
+    schema = Schema(C2)
+    schema(C2())
+
+    try:
+        schema(None)
+    except MultipleInvalid as e:
+        assert_equal(str(e), "expected C2")
+        assert_equal(len(e.errors), 1)
+        assert_equal(type(e.errors[0]), TypeInvalid)
+    else:
+        assert False, "Did not raise Invalid"
+
+
 def test_email_validation():
     """ test with valid email """
     schema = Schema({"email": Email()})
@@ -375,6 +411,69 @@ def test_subschema_extension():
     assert_equal(extended.schema, {'a': {'b': str, 'c': float, 'e': int}, 'd': str})
 
 
+def test_equality():
+    assert_equal(Schema('foo'), Schema('foo'))
+
+    assert_equal(Schema(['foo', 'bar', 'baz']),
+                 Schema(['foo', 'bar', 'baz']))
+
+    # Ensure two Schemas w/ two equivalent dicts initialized in a different
+    # order are considered equal.
+    dict_a = {}
+    dict_a['foo'] = 1
+    dict_a['bar'] = 2
+    dict_a['baz'] = 3
+
+    dict_b = {}
+    dict_b['baz'] = 3
+    dict_b['bar'] = 2
+    dict_b['foo'] = 1
+
+    assert_equal(Schema(dict_a), Schema(dict_b))
+
+
+def test_equality_negative():
+    """Verify that Schema objects are not equal to string representations"""
+    assert_false(Schema('foo') == 'foo')
+
+    assert_false(Schema(['foo', 'bar']) == "['foo', 'bar']")
+    assert_false(Schema(['foo', 'bar']) == Schema("['foo', 'bar']"))
+
+    assert_false(Schema({'foo': 1, 'bar': 2}) == "{'foo': 1, 'bar': 2}")
+    assert_false(Schema({'foo': 1, 'bar': 2}) == Schema("{'foo': 1, 'bar': 2}"))
+
+
+def test_inequality():
+    assert_true(Schema('foo') != 'foo')
+
+    assert_true(Schema(['foo', 'bar']) != "['foo', 'bar']")
+    assert_true(Schema(['foo', 'bar']) != Schema("['foo', 'bar']"))
+
+    assert_true(Schema({'foo': 1, 'bar': 2}) != "{'foo': 1, 'bar': 2}")
+    assert_true(Schema({'foo': 1, 'bar': 2}) != Schema("{'foo': 1, 'bar': 2}"))
+
+
+def test_inequality_negative():
+    assert_false(Schema('foo') != Schema('foo'))
+
+    assert_false(Schema(['foo', 'bar', 'baz']) !=
+                 Schema(['foo', 'bar', 'baz']))
+
+    # Ensure two Schemas w/ two equivalent dicts initialized in a different
+    # order are considered equal.
+    dict_a = {}
+    dict_a['foo'] = 1
+    dict_a['bar'] = 2
+    dict_a['baz'] = 3
+
+    dict_b = {}
+    dict_b['baz'] = 3
+    dict_b['bar'] = 2
+    dict_b['foo'] = 1
+
+    assert_false(Schema(dict_a) != Schema(dict_b))
+
+
 def test_repr():
     """Verify that __repr__ returns valid Python expressions"""
     match = Match('a pattern', msg='message')
@@ -393,7 +492,7 @@ def test_repr():
     )
     assert_equal(repr(coerce_), "Coerce(int, msg='moo')")
     assert_equal(repr(all_), "All('10', Coerce(int, msg=None), msg='all msg')")
-    assert_equal(repr(maybe_int), "Maybe(%s)" % str(int))
+    assert_equal(repr(maybe_int), "Any(None, %s, msg=None)" % str(int))
 
 
 def test_list_validation_messages():
@@ -514,14 +613,16 @@ def test_unordered():
 
 
 def test_maybe():
-    assert_raises(TypeError, Maybe, lambda x: x)
-
     s = Schema(Maybe(int))
     assert s(1) == 1
     assert s(None) is None
-
     assert_raises(Invalid, s, 'foo')
 
+    s = Schema(Maybe({str: Coerce(int)}))
+    assert s({'foo': '100'}) == {'foo': 100}
+    assert s(None) is None
+    assert_raises(Invalid, s, {'foo': 'bar'})
+
 
 def test_empty_list_as_exact():
     s = Schema([])
@@ -529,40 +630,6 @@ def test_empty_list_as_exact():
     s([])
 
 
-def test_empty_dict_as_exact():
-    # {} always evaluates as {}
-    s = Schema({})
-    assert_raises(Invalid, s, {'extra': 1})
-    s = Schema({}, extra=ALLOW_EXTRA)  # this should not be used
-    assert_raises(Invalid, s, {'extra': 1})
-
-    # {...} evaluates as Schema({...})
-    s = Schema({'foo': int})
-    assert_raises(Invalid, s, {'foo': 1, 'extra': 1})
-    s = Schema({'foo': int}, extra=ALLOW_EXTRA)
-    s({'foo': 1, 'extra': 1})
-
-    # dict matches {} or {...}
-    s = Schema(dict)
-    s({'extra': 1})
-    s({})
-    s = Schema(dict, extra=PREVENT_EXTRA)
-    s({'extra': 1})
-    s({})
-
-    # nested {} evaluate as {}
-    s = Schema({
-        'inner': {}
-    }, extra=ALLOW_EXTRA)
-    assert_raises(Invalid, s, {'inner': {'extra': 1}})
-    s({})
-    s = Schema({
-        'inner': Schema({}, extra=ALLOW_EXTRA)
-    })
-    assert_raises(Invalid, s, {'inner': {'extra': 1}})
-    s({})
-
-
 def test_schema_decorator_match_with_args():
     @validate(int)
     def fn(arg):
@@ -643,6 +710,38 @@ def test_schema_decorator_return_only_unmatch():
     assert_raises(Invalid, fn, 1)
 
 
+def test_schema_decorator_partial_match_called_with_args():
+    @validate(arg1=int)
+    def fn(arg1, arg2):
+        return arg1
+
+    fn(1, "foo")
+
+
+def test_schema_decorator_partial_unmatch_called_with_args():
+    @validate(arg1=int)
+    def fn(arg1, arg2):
+        return arg1
+
+    assert_raises(Invalid, fn, "bar", "foo")
+
+
+def test_schema_decorator_partial_match_called_with_kwargs():
+    @validate(arg2=int)
+    def fn(arg1, arg2):
+        return arg1
+
+    fn(arg1="foo", arg2=1)
+
+
+def test_schema_decorator_partial_unmatch_called_with_kwargs():
+    @validate(arg2=int)
+    def fn(arg1, arg2):
+        return arg1
+
+    assert_raises(Invalid, fn, arg1=1, arg2="foo")
+
+
 def test_unicode_as_key():
     if sys.version_info >= (3,):
         text_type = str
@@ -793,6 +892,79 @@ def test_marker_hashable():
     assert_equal(definition.get('j'), None)
 
 
+def test_schema_infer():
+    schema = Schema.infer({
+        'str': 'foo',
+        'bool': True,
+        'int': 42,
+        'float': 3.14
+    })
+    assert_equal(schema, Schema({
+        Required('str'): str,
+        Required('bool'): bool,
+        Required('int'): int,
+        Required('float'): float
+    }))
+
+
+def test_schema_infer_dict():
+    schema = Schema.infer({
+        'a': {
+            'b': {
+                'c': 'foo'
+            }
+        }
+    })
+
+    assert_equal(schema, Schema({
+        Required('a'): {
+            Required('b'): {
+                Required('c'): str
+            }
+        }
+    }))
+
+
+def test_schema_infer_list():
+    schema = Schema.infer({
+        'list': ['foo', True, 42, 3.14]
+    })
+
+    assert_equal(schema, Schema({
+        Required('list'): [str, bool, int, float]
+    }))
+
+
+def test_schema_infer_scalar():
+    assert_equal(Schema.infer('foo'), Schema(str))
+    assert_equal(Schema.infer(True), Schema(bool))
+    assert_equal(Schema.infer(42), Schema(int))
+    assert_equal(Schema.infer(3.14), Schema(float))
+    assert_equal(Schema.infer({}), Schema(dict))
+    assert_equal(Schema.infer([]), Schema(list))
+
+
+def test_schema_infer_accepts_kwargs():
+    schema = Schema.infer({
+        'str': 'foo',
+        'bool': True
+    }, required=False, extra=True)
+
+    # Subset of schema should be acceptable thanks to required=False.
+    schema({'bool': False})
+
+    # Keys that are in schema should still match required types.
+    try:
+        schema({'str': 42})
+    except Invalid:
+        pass
+    else:
+        assert False, 'Did not raise Invalid for Number'
+
+    # Extra fields should be acceptable thanks to extra=True.
+    schema({'str': 'bar', 'int': 42})
+
+
 def test_validation_performance():
     """
     This test comes to make sure the validation complexity of dictionaries is done in a linear time.
@@ -816,7 +988,7 @@ def test_validation_performance():
     for i in range(num_of_keys):
         schema_dict[CounterMarker(str(i))] = str
         data[str(i)] = str(i)
-        data_extra_keys[str(i*2)] = str(i)  # half of the keys are present, and half aren't
+        data_extra_keys[str(i * 2)] = str(i)  # half of the keys are present, and half aren't
 
     schema = Schema(schema_dict, extra=ALLOW_EXTRA)
 
@@ -828,3 +1000,144 @@ def test_validation_performance():
     schema(data_extra_keys)
 
     assert counter[0] <= num_of_keys, "Validation complexity is not linear! %s > %s" % (counter[0], num_of_keys)
+
+
+def test_IsDir():
+    schema = Schema(IsDir())
+    assert_raises(MultipleInvalid, schema, 3)
+    schema(os.path.dirname(os.path.abspath(__file__)))
+
+
+def test_IsFile():
+    schema = Schema(IsFile())
+    assert_raises(MultipleInvalid, schema, 3)
+    schema(os.path.abspath(__file__))
+
+
+def test_PathExists():
+    schema = Schema(PathExists())
+    assert_raises(MultipleInvalid, schema, 3)
+    schema(os.path.abspath(__file__))
+
+
+def test_description():
+    marker = Marker(Schema(str), description='Hello')
+    assert marker.description == 'Hello'
+
+    optional = Optional('key', description='Hello')
+    assert optional.description == 'Hello'
+
+    exclusive = Exclusive('alpha', 'angles', description='Hello')
+    assert exclusive.description == 'Hello'
+
+    required = Required('key', description='Hello')
+    assert required.description == 'Hello'
+
+
+def test_SomeOf_min_validation():
+    validator = All(Length(min=8), SomeOf(
+        min_valid=3,
+        validators=[Match(r'.*[A-Z]', 'no uppercase letters'),
+                    Match(r'.*[a-z]', 'no lowercase letters'),
+                    Match(r'.*[0-9]', 'no numbers'),
+                    Match(r'.*[$@$!%*#?&^:;/<,>|{}()\-\'._+=]', 'no symbols')]))
+
+    validator('ffe532A1!')
+    with raises(MultipleInvalid, 'length of value must be at least 8'):
+        validator('a')
+
+    with raises(MultipleInvalid, 'no uppercase letters, no lowercase letters'):
+        validator('wqs2!#s111')
+
+    with raises(MultipleInvalid, 'no lowercase letters, no symbols'):
+        validator('3A34SDEF5')
+
+
+def test_SomeOf_max_validation():
+    validator = SomeOf(
+        max_valid=2,
+        validators=[Match(r'.*[A-Z]', 'no uppercase letters'),
+                    Match(r'.*[a-z]', 'no lowercase letters'),
+                    Match(r'.*[0-9]', 'no numbers')],
+        msg='max validation test failed')
+
+    validator('Aa')
+    with raises(TooManyValid, 'max validation test failed'):
+        validator('Aa1')
+
+
+def test_self_validation():
+    schema = Schema({"number": int,
+                     "follow": Self})
+    try:
+        schema({"number": "abc"})
+    except MultipleInvalid:
+        pass
+    else:
+        assert False, "Did not raise Invalid"
+    try:
+        schema({"follow": {"number": '123456.712'}})
+    except MultipleInvalid:
+        pass
+    else:
+        assert False, "Did not raise Invalid"
+    schema({"follow": {"number": 123456}})
+    schema({"follow": {"follow": {"number": 123456}}})
+
+
+def test_self_any():
+    schema = Schema({"number": int,
+                     "follow": Any(Self, "stop")})
+    try:
+        schema({"number": "abc"})
+    except MultipleInvalid:
+        pass
+    else:
+        assert False, "Did not raise Invalid"
+    try:
+        schema({"follow": {"number": '123456.712'}})
+    except MultipleInvalid:
+        pass
+    else:
+        assert False, "Did not raise Invalid"
+    schema({"follow": {"number": 123456}})
+    schema({"follow": {"follow": {"number": 123456}}})
+    schema({"follow": {"follow": {"number": 123456, "follow": "stop"}}})
+
+
+def test_self_all():
+    schema = Schema({"number": int,
+                     "follow": All(Self,
+                                   Schema({"extra_number": int},
+                                          extra=ALLOW_EXTRA))},
+                    extra=ALLOW_EXTRA)
+    try:
+        schema({"number": "abc"})
+    except MultipleInvalid:
+        pass
+    else:
+        assert False, "Did not raise Invalid"
+    try:
+        schema({"follow": {"number": '123456.712'}})
+    except MultipleInvalid:
+        pass
+    else:
+        assert False, "Did not raise Invalid"
+    schema({"follow": {"number": 123456}})
+    schema({"follow": {"follow": {"number": 123456}}})
+    schema({"follow": {"number": 123456, "extra_number": 123}})
+    try:
+        schema({"follow": {"number": 123456, "extra_number": "123"}})
+    except MultipleInvalid:
+        pass
+    else:
+        assert False, "Did not raise Invalid"
+
+
+def test_SomeOf_on_bounds_assertion():
+    with raises(AssertionError, 'when using "SomeOf" you should specify at least one of min_valid and max_valid'):
+        SomeOf(validators=[])
+
+
+def test_comparing_voluptuous_object_to_str():
+    assert_true(Optional('Classification') < 'Name')
